from typing import List, Literal, Optional, Union, cast
from warnings import warn

import httpx
from pydantic import BaseModel

import langwatch
import langwatch.tracer
from langwatch.tracer import ContextSpan, get_current_tracer, _local_context
from langwatch.types import GuardrailResult, RAGChunk, TypedValueGuardrailResult


class Money(BaseModel):
    currency: str
    amount: float


class GuardrailResultModel(BaseModel):
    status: Literal["processed", "skipped", "error"]
    passed: bool = True
    score: Optional[float] = None
    details: Optional[str] = None
    cost: Optional[Money] = None


def evaluate(
    slug: str,
    input: Optional[str] = None,
    output: Optional[str] = None,
    contexts: List[RAGChunk] = [],
):
    with langwatch.tracer.create_span(name=slug, type="guardrail") as span:
        request_params = prepare_data(slug, input, output, contexts, span=span)
        try:
            with httpx.Client() as client:
                response = client.post(**request_params)
                response.raise_for_status()
        except Exception as e:
            return handle_response(
                {
                    "status": "error",
                    "message": str(e),
                    "passed": True,
                },
                span,
            )

        return handle_response(response.json(), span)


async def async_evaluate(
    slug: str,
    input: Optional[str] = None,
    output: Optional[str] = None,
    contexts: List[RAGChunk] = [],
):
    current_span = getattr(_local_context, "current_span", None)
    with langwatch.tracer.create_span(name=slug, type="guardrail") as span:
        # hack: avoid nesting async evaluate spans when they are called in parallel
        _local_context.current_span = current_span

        request_params = prepare_data(slug, input, output, contexts, span=span)
        try:
            async with httpx.AsyncClient() as client:
                response = await client.post(**request_params)
                response.raise_for_status()
        except Exception as e:
            return handle_response(
                {
                    "status": "error",
                    "message": str(e),
                    "passed": True,
                },
                span,
            )

        return handle_response(response.json(), span)


def prepare_data(
    slug: str,
    input: Optional[str],
    output: Optional[str],
    contexts: List[RAGChunk],
    span: ContextSpan,
):
    current_tracer = get_current_tracer()
    data = {}
    if input:
        data["input"] = input
    if output:
        data["output"] = output
    if contexts and len(contexts) > 0:
        data["contexts"] = contexts
    span.input = data

    return {
        "url": langwatch.endpoint + f"/api/guardrails/{slug}/evaluate",
        "json": {
            "trace_id": current_tracer.trace_id if current_tracer else None,
            "data": data,
        },
        "headers": {"X-Auth-Token": str(langwatch.api_key)},
    }


def handle_response(
    response: dict,
    span: ContextSpan,
):
    result = GuardrailResultModel.model_validate(response)
    if result.status == "error":
        result.details = response.get("message", "")
    span.output = TypedValueGuardrailResult(
        type="guardrail_result", value=cast(GuardrailResult, result.model_dump())
    )
    return result



import functools
import os
import threading
from concurrent.futures import Future, ThreadPoolExecutor
import time
from typing import Any, Dict, List, Optional, TypeVar
from warnings import warn

import nanoid
import requests
from langwatch.types import (
    BaseSpan,
    ErrorCapture,
    RAGChunk,
    RAGSpan,
    Span,
    SpanTimestamps,
    SpanTypes,
    CollectorRESTParams,
    TraceMetadata,
)
from langwatch.utils import (
    autoconvert_typed_values,
    capture_exception,
    milliseconds_timestamp,
)
from retry import retry

import langwatch
import contextvars  # Import contextvars module

class ContextSpan:
    span_id: str
    parent: Optional["ContextSpan"] = None
    name: Optional[str]
    type: SpanTypes
    input: Any
    output: Optional[Any] = None
    started_at: int

    def __init__(
        self,
        span_id: str,
        name: Optional[str],
        type: SpanTypes = "span",
        input: Any = None,
    ) -> None:
        self.span_id = span_id
        self.name = name
        self.type = type
        self.input = input

        current_tracer = current_tracer_var.get()

        if current_tracer:
            if current_tracer.current_span:
                self.parent = current_tracer.current_span
            current_tracer.current_span = self
        else:
            warn("No current tracer found, some spans will not be sent to LangWatch")

        self.started_at = milliseconds_timestamp()

    def __enter__(self):
        return self

    def __exit__(self, _exc_type, exc_value: Optional[BaseException], _exc_traceback):
        error: Optional[ErrorCapture] = (
            capture_exception(exc_value) if exc_value else None
        )
        finished_at = milliseconds_timestamp()

        current_tracer = current_tracer_var.get()
        if current_tracer:
            span_id = self.span_id  # TODO: test?
            current_tracer.append_span(
                self.create_span(span_id, current_tracer, error, finished_at)
            )

            current_tracer.current_span = self.parent

    def create_span(
        self,
        span_id: str,
        context_tracer: Any,
        error: Optional[ErrorCapture],
        finished_at: int,
    ) -> BaseSpan:
        return BaseSpan(
            type=self.type,
            name=self.name,
            span_id=span_id,
            parent_id=self.parent.span_id if self.parent else None,
            trace_id=context_tracer.trace_id,
            input=autoconvert_typed_values(self.input) if self.input else None,
            outputs=([autoconvert_typed_values(self.output)] if self.output else []),
            error=error,
            metrics=None,
            timestamps=SpanTimestamps(
                started_at=self.started_at, finished_at=finished_at
            ),
        )


class ContextRAGSpan(ContextSpan):
    contexts: List[RAGChunk]

    def __init__(
        self,
        id: str,
        name: Optional[str],
        input: Any = None,
        contexts: List[RAGChunk] = [],
    ) -> None:
        super().__init__(id, name, type="rag", input=input)
        self.contexts = contexts

    def create_span(
        self,
        id: str,
        context_tracer: Any,
        error: Optional[ErrorCapture],
        finished_at: int,
    ) -> RAGSpan:
        span: Any = super().create_span(id, context_tracer, error, finished_at)
        rag_span: RAGSpan = {**span, "type": "rag", "contexts": self.contexts}

        return rag_span


def create_span(
    name: Optional[str] = None, type: SpanTypes = "span", input: Any = None
):
    return ContextSpan(
        span_id=f"span_{nanoid.generate()}", name=name, type=type, input=input
    )


def capture_rag(
    contexts: List[RAGChunk],
    input: Optional[str] = None,
    name: str = "RetrievalAugmentedGeneration",
):
    return ContextRAGSpan(
        id=f"span_{nanoid.generate()}", name=name, input=input, contexts=contexts
    )


def span(name: Optional[str] = None, type: SpanTypes = "span"):
    def _span(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            all_args = (
                {str(index): item for index, item in enumerate(args)} if args else {}
            )
            if kwargs:
                all_args.update(kwargs)

            with create_span(
                name=(name or func.__name__), type=type, input=all_args
            ) as span:
                output = func(*args, **kwargs)
                span.output = output
                return output

        return wrapper

    return _span


def get_current_tracer():
    return current_tracer_var.get()


executor = ThreadPoolExecutor(max_workers=10)


class BaseContextTracer:
    sent_once = False
    scheduled_send: Optional[Future[None]] = None
    current_span: Optional[ContextSpan] = None

    def __init__(
        self,
        trace_id: Optional[str],
        metadata: Optional[TraceMetadata],
    ):
        self.spans: Dict[str, Span] = {}
        self.trace_id = trace_id or f"trace_{nanoid.generate()}"
        self.metadata = metadata

    def __enter__(self):
        self.token = current_tracer_var.set(self)
        return self

    def __exit__(self, _type, _value, _traceback):
        self.delayed_send_spans()
        current_tracer_var.reset(self.token)

    def delayed_send_spans(self):
        self._add_finished_at_to_missing_spans()

        if "PYTEST_CURRENT_TEST" in os.environ:
            # Keep on the same thread for tests
            self.send_spans_sync()
            return

        def run_in_thread():
            time.sleep(1)  # wait for other spans to be added
            self.sent_once = True
            self.send_spans_sync()

        if self.scheduled_send and not self.scheduled_send.done():
            self.scheduled_send.cancel()

        self.scheduled_send = executor.submit(run_in_thread)

    def send_spans_sync(self):
        send_spans(
            CollectorRESTParams(
                trace_id=self.trace_id,
                metadata=self.metadata,
                spans=list(self.spans.values()),
            )
        )

    def append_span(self, span: Span):
        span["span_id"] = span.get("span_id", f"span_{nanoid.generate()}")
        self.spans[span["span_id"]] = span
        if self.sent_once:
            self.delayed_send_spans()  # send again if needed

    def get_parent_id(self):
        if self.current_span:
            return self.current_span.span_id
        return None

    # Some spans get interrupted in the middle, for example by an exception, and we might end up never tagging their finish timestamp, so we do it here as a fallback
    def _add_finished_at_to_missing_spans(self):
        for span in self.spans.values():
            if "timestamps" in span and (
                "finished_at" not in span["timestamps"]
                or span["timestamps"]["finished_at"] == None
            ):
                span["timestamps"]["finished_at"] = milliseconds_timestamp()


current_tracer_var = contextvars.ContextVar[Optional[BaseContextTracer]]("current_tracer", default=None)


@retry(tries=5, delay=0.5, backoff=3)
def send_spans(data: CollectorRESTParams):
    if len(data["spans"]) == 0:
        return
    if not langwatch.api_key:
        warn(
            "LANGWATCH_API_KEY is not set, LLMs traces will not be sent, go to https://langwatch.ai to set it up"
        )
        return
    response = requests.post(
        langwatch.endpoint + "/api/collector",
        json=data,
        headers={"X-Auth-Token": str(langwatch.api_key)},
    )
    if response.status_code == 429:
        json = response.json()
        if "message" in json and "ERR_PLAN_LIMIT" in json["message"]:
            warn(json["message"])
        else:
            warn("Rate limit exceeded, dropping message from being sent to LangWatch")
    else:
        response.raise_for_status()



from typing import Any, Dict, List, Literal, Optional, TypedDict, Union

from pydantic import BaseModel, Field


ChatRole = Literal[
    "system", "user", "assistant", "function", "tool", "guardrail", "unknown"
]


class FunctionCall(TypedDict, total=False):
    name: str
    arguments: str


class ToolCall(TypedDict):
    id: str
    type: str
    function: FunctionCall


class ChatMessage(TypedDict, total=False):
    role: ChatRole
    content: Optional[str]
    function_call: Optional[FunctionCall]
    tool_calls: Optional[List[ToolCall]]


class TypedValueChatMessages(TypedDict):
    type: Literal["chat_messages"]
    value: List[ChatMessage]


class TypedValueText(TypedDict):
    type: Literal["text"]
    value: str


class GuardrailResult(TypedDict, total=False):
    status: Literal["processed", "skipped", "error"]
    passed: bool
    score: Optional[float]
    details: Optional[str]


class TypedValueGuardrailResult(TypedDict):
    type: Literal["guardrail_result"]
    value: GuardrailResult


class TypedValueRaw(TypedDict):
    type: Literal["raw"]
    value: str


JSONSerializable = Union[str, int, float, bool, None, Dict[str, Any], List[Any]]


class TypedValueJson(TypedDict):
    type: Literal["json"]
    value: JSONSerializable


class ErrorCapture(TypedDict):
    message: str
    stacktrace: List[str]


SpanInputOutput = Union[
    TypedValueText,
    TypedValueChatMessages,
    TypedValueJson,
    TypedValueGuardrailResult,
    TypedValueRaw,
]


class SpanMetrics(TypedDict, total=False):
    prompt_tokens: Optional[int]
    completion_tokens: Optional[int]


class SpanParams(TypedDict, total=False):
    temperature: float
    stream: bool
    functions: Optional[List[Dict[str, Any]]]
    tools: Optional[List[Dict[str, Any]]]
    tool_choice: Optional[str]


class SpanTimestamps(TypedDict, total=False):
    started_at: int
    first_token_at: Optional[int]
    finished_at: int


SpanTypes = Literal["span", "llm", "chain", "tool", "agent", "guardrail", "rag"]


class BaseSpan(TypedDict):
    type: SpanTypes
    name: Optional[str]
    span_id: str
    parent_id: Optional[str]
    trace_id: str
    input: Optional[SpanInputOutput]
    outputs: List[SpanInputOutput]
    error: Optional[ErrorCapture]
    metrics: Optional[SpanMetrics]
    timestamps: SpanTimestamps


class LLMSpan(TypedDict, total=False):
    type: Literal["llm"]
    span_id: str
    parent_id: Optional[str]
    trace_id: str
    vendor: str
    model: str
    input: SpanInputOutput
    outputs: List[SpanInputOutput]
    error: Optional[ErrorCapture]
    params: SpanParams
    metrics: Optional[SpanMetrics]
    timestamps: SpanTimestamps


class RAGChunk(TypedDict, total=False):
    document_id: Optional[str]
    chunk_id: Optional[str]
    content: Union[str, dict, list]


class RAGSpan(TypedDict, total=False):
    type: Literal["rag"]
    name: Optional[str]
    span_id: str
    parent_id: Optional[str]
    trace_id: str
    input: Optional[SpanInputOutput]
    outputs: List[SpanInputOutput]
    error: Optional[ErrorCapture]
    metrics: Optional[SpanMetrics]
    timestamps: SpanTimestamps
    contexts: List[RAGChunk]


Span = Union[LLMSpan, RAGSpan, BaseSpan]


class Experiment(TypedDict):
    experiment_id: str
    variant: int


class TraceMetadata(TypedDict, total=False):
    user_id: Optional[str]
    thread_id: Optional[str]
    customer_id: Optional[str]
    labels: List[str]
    experiments: List[Experiment]


class CollectorRESTParams(TypedDict):
    trace_id: str
    metadata: Optional[TraceMetadata]
    spans: List[Span]



import asyncio
from concurrent.futures import Future, ThreadPoolExecutor, as_completed
from typing import (
    Any,
    Callable,
    Coroutine,
    List,
    Literal,
    Optional,
    Tuple,
    TypedDict,
    Union,
)
from pydantic import BaseModel, Field
import nanoid
import langwatch
import httpx
from tqdm import tqdm
import pandas as pd


class Money(BaseModel):
    currency: str
    amount: float


class EvaluationResult(BaseModel):
    status: Literal["processed"] = "processed"
    score: float = Field(description="No description provided")
    passed: Optional[bool] = None
    details: Optional[str] = Field(
        default=None, description="Short human-readable description of the result"
    )
    cost: Optional[Money] = None


class EvaluationResultSkipped(BaseModel):
    status: Literal["skipped"] = "skipped"
    details: Optional[str] = None


class EvaluationResultError(BaseModel):
    status: Literal["error"] = "error"
    error_type: str = Field(description="The type of the exception")
    message: str = Field(description="Error message")
    traceback: List[str] = Field(description="Traceback information for debugging")


SingleEvaluationResult = (
    EvaluationResult | EvaluationResultSkipped | EvaluationResultError
)


class DatasetEntry(BaseModel):
    input: str
    expected_output: str


class DatasetEntryWithOutput(DatasetEntry):
    output: str


class DatasetRecord(BaseModel):
    id: str
    entry: DatasetEntry


class BatchEvaluationResultRecord(BaseModel):
    entry: DatasetEntryWithOutput
    results: list[Tuple[str, SingleEvaluationResult]]


class BatchEvaluationRun(BaseModel):
    model_config = {"arbitrary_types_allowed": True}

    list: list[BatchEvaluationResultRecord]
    df: pd.DataFrame


class BatchEvaluation:
    def __init__(
        self,
        dataset: str,
        evaluations: list[str],
        callback: Callable[[DatasetEntry], Union[str, dict[str, str]]],
        generations="one-shot",
        max_workers=4,
    ):
        self.dataset = dataset
        self.evaluations = evaluations
        self.callback = callback
        self.generations = generations
        self.max_workers = max_workers

    def run(self):
        print("Starting batch evaluation...")
        batchId = f"batch_{nanoid.generate()}"
        dataset = get_dataset(self.dataset)

        if dataset is None:
            raise Exception(f"Dataset {self.dataset} not found.")

        if len(dataset) == 0:
            raise Exception(f"Dataset {self.dataset} is empty.")

        results: list[BatchEvaluationResultRecord] = []

        with ThreadPoolExecutor(max_workers=self.max_workers) as executor:
            futures: List[Future] = [
                executor.submit(self.run_record, record, batchId) for record in dataset
            ]
            for future in tqdm(as_completed(futures), total=len(futures)):
                results.append(future.result())

        return BatchEvaluationRun(
            list=results,
            df=self._results_to_pandas(results),
        )

    def _results_to_pandas(self, results: list[BatchEvaluationResultRecord]):
        results_df = []
        for result in results:
            result_dict: dict[str, Any] = {
                "input": result.entry.input,
                "expected_output": result.entry.expected_output,
                "output": result.entry.output,
            }
            for evaluation_name, evaluation_result in result.results:
                if evaluation_result.status == "processed":
                    result_dict[evaluation_name] = (
                        evaluation_result.score
                        if evaluation_result.passed is None
                        else evaluation_result.passed
                    )
                else:
                    result_dict[evaluation_name] = evaluation_result.status
            results_df.append(result_dict)

        return pd.DataFrame(results_df)

    def run_record(self, record: DatasetRecord, batchId: str):
        entry = record.entry

        callbackResponse = self.callback(entry)
        entry_with_output = DatasetEntryWithOutput(
            input=entry.input,
            expected_output=entry.expected_output,
            output=(
                callbackResponse
                if isinstance(callbackResponse, str)
                else callbackResponse["output"]
            ),
        )

        coroutines: list[Coroutine[Tuple[str, SingleEvaluationResult], Any, Any]] = []
        for evaluation in self.evaluations:
            coroutines.append(
                run_evaluation(entry_with_output, evaluation, batchId, self.dataset)
            )

        async def gather_results(futures):
            return await asyncio.gather(*futures)

        evaluation_results: list[Tuple[str, SingleEvaluationResult]] = asyncio.run(
            gather_results(coroutines)
        )

        return BatchEvaluationResultRecord(
            entry=entry_with_output, results=evaluation_results
        )


async def run_evaluation(
    data: DatasetEntryWithOutput, evaluation: str, batchId: str, datasetSlug: str
) -> Tuple[str, SingleEvaluationResult]:
    try:
        json_data = {
            "data": data.model_dump(),
            "evaluation": evaluation,
            "batchId": batchId,
            "datasetSlug": datasetSlug,
        }

        request_params = {
            "url": langwatch.endpoint + f"/api/dataset/evaluate",
            "headers": {"X-Auth-Token": str(langwatch.api_key)},
            "json": json_data,
        }

        async with httpx.AsyncClient(timeout=60) as client:
            response = await client.post(**request_params)
            response.raise_for_status()

        result = response.json()
        if result["status"] == "processed":
            return evaluation, EvaluationResult.model_validate(result)
        elif result["status"] == "skipped":
            return evaluation, EvaluationResultSkipped.model_validate(result)
        else:
            return evaluation, EvaluationResultError.model_validate(result)

    except httpx.HTTPStatusError as e:
        if e.response.status_code // 100 == 4:
            raise Exception(f"HTTP error: {e.response.text}")
        else:
            return evaluation, EvaluationResultError(
                error_type="HTTPStatusError",
                message=f"HTTP error {e.response.status_code}: {e.response.text}",
                traceback=[],
            )

    except Exception as e:
        return evaluation, EvaluationResultError(
            error_type="Exception",
            message=str(e),
            traceback=[],
        )


def get_dataset(
    slug: str,
) -> list[DatasetRecord]:
    request_params = {
        "url": langwatch.endpoint + f"/api/dataset/{slug}",
        "headers": {"X-Auth-Token": str(langwatch.api_key)},
    }
    with httpx.Client() as client:
        response = client.get(**request_params)
        response.raise_for_status()

    result = response.json()

    if "status" in result and result["status"] == "error":
        # If the response contains a status key and its value is "error"
        error_message = result.get("message", "Unknown error")
        # Get the error message if available, otherwise set it to "Unknown error"
        raise Exception(f"Error: {error_message}")
    else:
        # If the response does not contain a status key or its value is not "error"
        res = result.get("data", None)
        # parse to pydantic
        return [DatasetRecord.model_validate(record) for record in res]
